import struct

import pytest

from threedi_schema import ModelSchema, ThreediDatabase

# from threedi_schema.migrations.versions import 0213_connected_points_to_breaches
module_name = "0213_connected_points_to_breaches"
migration_213 = getattr(
    __import__("threedi_schema.migrations.versions", fromlist=[module_name]),
    module_name,
)

CalculationPoint = migration_213.CalculationPoint
ConnectedPoint = migration_213.ConnectedPoint
Levee = migration_213.Levee
PotentialBreach = migration_213.PotentialBreach
ConnectionNode = migration_213.ConnectionNode
Manhole = migration_213.Manhole
Channel = migration_213.Channel


@pytest.fixture
def sqlite_v212():
    """An in-memory database with schema version 212"""
    db = ThreediDatabase("")
    ModelSchema(db).upgrade("0212", backup=False)
    return db


@pytest.fixture
def session(sqlite_v212):
    return sqlite_v212.get_session()


def parse_hexewkb(ewkb):
    """Extract 4 line coordinates from an ewkb"""
    return struct.unpack("4d", bytes.fromhex(ewkb)[-32:])


GEOM1 = "SRID=4326;POINT (0 0)"
GEOM2 = "SRID=4326;POINT (0 1)"
CHANNEL = "SRID=4326;LINESTRING(0 0,10 0)"
CHANNEL_INV = "SRID=4326;LINESTRING(10 0,0 0)"
LINE = "0102000020E610000002000000000000000000000000000000000000000000000000000000000000000000F03F"


@pytest.mark.parametrize(
    "objs",
    [
        [CalculationPoint(id=1, the_geom=GEOM1)],
        [ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=None)],
        [ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=1)],
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=1),
        ],
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=1),
            ConnectedPoint(id=2, the_geom=GEOM1, calculation_pnt_id=1),
        ],
    ],
)
def test_clean_connected_points(session, objs):
    session.add_all(objs)
    session.flush()
    migration_213.clean_connected_points(session)
    assert session.query(CalculationPoint).count() == 0
    assert session.query(ConnectedPoint).count() == 0
#
#
# @pytest.mark.parametrize(
#     "objs",
#     [
#         [
#             CalculationPoint(id=1, the_geom=GEOM1),
#             ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#         ],
#         [
#             CalculationPoint(id=1, the_geom=GEOM1),
#             ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#             ConnectedPoint(id=2, the_geom=GEOM2, calculation_pnt_id=1),
#         ],
#         [
#             CalculationPoint(id=1, the_geom=GEOM1),
#             ConnectedPoint(
#                 id=1, the_geom=GEOM2, calculation_pnt_id=1, exchange_level=1.0
#             ),
#         ],
#         [
#             CalculationPoint(id=1, the_geom=GEOM1),
#             ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1, levee_id=1),
#         ],
#     ],
# )
# def test_clean_connected_points_keep(session, objs):
#     session.add_all(objs)
#     session.flush()
#     migration_213.clean_connected_points(session)
#
#     actual = (
#         session.query(CalculationPoint).count() + session.query(ConnectedPoint).count()
#     )
#     assert actual == len(objs)
#
#
# def todict(x):
#     return {col.name: getattr(x, col.name) for col in x.__table__.columns}
#
#
# def assert_sqlalchemy_objects_equal(a, b):
#     assert a.__class__ is b.__class__
#     if a is None:
#         return
#     assert todict(a) == todict(b)
#
#
# @pytest.mark.parametrize(
#     "objs,expected",
#     [
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#                 Channel(id=4, the_geom=CHANNEL),
#             ],
#             PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#4#v2_channel#1"),
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
#                 ConnectedPoint(
#                     id=1, the_geom=GEOM2, calculation_pnt_id=1, exchange_level=1.1
#                 ),
#                 Channel(id=4, the_geom=CHANNEL),
#             ],
#             PotentialBreach(
#                 channel_id=4,
#                 the_geom=LINE,
#                 exchange_level=1.1,
#                 code="1#123#4#v2_channel#1",
#             ),
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1, levee_id=4),
#                 Levee(id=4, crest_level=1.1),
#                 Channel(id=4, the_geom=CHANNEL),
#             ],
#             PotentialBreach(
#                 channel_id=4,
#                 the_geom=LINE,
#                 exchange_level=1.1,
#                 code="1#123#4#v2_channel#1",
#             ),
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
#                 ConnectedPoint(
#                     id=1,
#                     the_geom=GEOM2,
#                     calculation_pnt_id=1,
#                     exchange_level=1.1,
#                     levee_id=4,
#                 ),
#                 Levee(id=4, crest_level=1.2, max_breach_depth=0.5, material=1),
#                 Channel(id=4, the_geom=CHANNEL),
#             ],
#             PotentialBreach(
#                 channel_id=4,
#                 the_geom=LINE,
#                 exchange_level=1.1,
#                 maximum_breach_depth=0.5,
#                 levee_material=1,
#                 code="1#123#4#v2_channel#1",
#             ),
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#                 Manhole(id=3, connection_node_id=6),
#                 ConnectionNode(id=6),
#                 Channel(
#                     id=4,
#                     the_geom=CHANNEL,
#                     connection_node_start_id=6,
#                     calculation_type=102,
#                 ),
#             ],
#             PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#                 Manhole(id=3, connection_node_id=6),
#                 ConnectionNode(id=6),
#                 Channel(
#                     id=4,
#                     the_geom=CHANNEL_INV,
#                     connection_node_end_id=6,
#                     calculation_type=102,
#                 ),
#             ],
#             PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#                 Manhole(id=3, connection_node_id=6),
#                 ConnectionNode(id=6),
#                 Channel(
#                     id=4,
#                     the_geom=CHANNEL,
#                     connection_node_start_id=6,
#                     calculation_type=105,
#                 ),
#             ],
#             PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#                 Manhole(id=3, connection_node_id=6),
#                 ConnectionNode(id=6),
#             ],
#             None,
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#                 Manhole(id=3, connection_node_id=6),
#                 ConnectionNode(id=6),
#                 Channel(id=4, connection_node_start_id=6, calculation_type=101),
#             ],
#             None,
#         ],
#         [
#             [
#                 CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
#                 ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
#                 Manhole(id=3, connection_node_id=6),
#                 ConnectionNode(id=6),
#                 Channel(id=3, connection_node_start_id=6, calculation_type=102),
#                 Channel(id=5, connection_node_start_id=6, calculation_type=105),
#                 Channel(
#                     id=4,
#                     the_geom=CHANNEL_INV,
#                     connection_node_end_id=6,
#                     calculation_type=105,
#                 ),
#             ],
#             PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
#         ],
#     ],
# )
# def test_to_potential_breach(session, objs, expected):
#     session.add_all(objs)
#     session.flush()
#     actual = migration_213.to_potential_breach(session, 1)
#
#     assert_sqlalchemy_objects_equal(actual, expected)
#
#
# @pytest.mark.parametrize(
#     "node_idx,calc_pnt_x,calc_pnt_y,x,y",
#     [
#         (0, 0, 0, 0, 0),
#         (-1, 0, 0, 10, 10),
#         (1, 0, 5, 0, 5),
#         (1, 0, 10, 0, 10),
#         (1, 0, 10 + 1e-8, 0, 10),
#         (2, 10 + 7e-8, 10 - 7e-8, 10, 10),
#     ],
# )
# def test_get_breach_line_geom(session, node_idx, calc_pnt_x, calc_pnt_y, x, y):
#     objs = [
#         ConnectedPoint(id=1, the_geom="SRID=4326;POINT (10 0)", calculation_pnt_id=2),
#         CalculationPoint(id=2, the_geom=f"SRID=4326;POINT({calc_pnt_x} {calc_pnt_y})"),
#         Channel(id=3, the_geom="SRID=4326;LINESTRING(0 0,0 10,10 10)"),
#     ]
#
#     session.add_all(objs)
#     session.flush()
#     geom = migration_213.get_breach_line_geom(session, 1, 3, node_idx)
#     x1, y1, x2, y2 = parse_hexewkb(geom)
#     assert x1 == x
#     assert y1 == y
#     assert x2 == 10.0
#     assert y2 == 0.0



import shutil
import sqlite3
import tempfile
from pathlib import Path

import pytest
from sqlalchemy import text

from threedi_schema import models, ModelSchema, ThreediDatabase
from threedi_schema.migrations.exceptions import InvalidSRIDException

data_dir = Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def sqlite_path():
    return data_dir.joinpath("test_crs_migation_28992.sqlite")


@pytest.fixture()
def db(tmp_path_factory, sqlite_path):
    tmp_sqlite = tmp_path_factory.mktemp("custom_dir").joinpath(sqlite_path.name)
    shutil.copy(sqlite_path, tmp_sqlite)
    return ThreediDatabase(tmp_sqlite)


# TODO - match other testing and use generic fixtures
@pytest.mark.parametrize("epsg_code", [
    999999, # non-existing
    2227, # projected / US survey foot
    4979, # not project
])
def test_check_valid_crs(db, epsg_code):
    session = db.get_session()
    # Update the epsg_code in ModelSettings
    session.execute(text(f"UPDATE model_settings SET epsg_code = {epsg_code}"))
    session.commit()
    with pytest.raises(InvalidSRIDException) as exc_info:
        db.schema.upgrade(backup=False)
    session.execute(text("UPDATE model_settings SET epsg_code = 28992"))


# TODO - match other testing and use generic fixtures
def test_migration(tmp_path_factory):
    # Ensure all geometries are transformed
    sqlite_path = data_dir.joinpath("v2_bergermeer_221.sqlite")
    tmp_sqlite = tmp_path_factory.mktemp("custom_dir").joinpath(sqlite_path.name)
    shutil.copy(sqlite_path, tmp_sqlite)
    schema = ModelSchema(ThreediDatabase(tmp_sqlite))
    schema.upgrade(backup=False)
    cursor = sqlite3.connect(schema.db.path).cursor()
    query = cursor.execute("SELECT srid FROM geometry_columns where f_table_name = 'geom'")
    epsg_matches = [int(item[0])==28992 for item in query.fetchall()]
    assert all(epsg_matches)





"""Reproject geometries to model CRS

Revision ID: 0230
Revises:
Create Date: 2024-11-12 12:30

"""
import sqlite3
import uuid

import sqlalchemy as sa
from alembic import op

from threedi_schema.migrations.exceptions import InvalidSRIDException

# revision identifiers, used by Alembic.
revision = "0230"
down_revision = "0229"
branch_labels = None
depends_on = None

GEOM_TABLES = ['boundary_condition_1d', 'boundary_condition_2d', 'channel', 'connection_node', 'measure_location',
               'measure_map', 'memory_control', 'table_control', 'cross_section_location', 'culvert',
               'dem_average_area', 'dry_weather_flow', 'dry_weather_flow_map', 'exchange_line', 'grid_refinement_line',
               'grid_refinement_area', 'lateral_1d', 'lateral_2d', 'obstacle', 'orifice', 'pipe', 'potential_breach',
               'pump', 'pump_map', 'surface', 'surface_map', 'weir', 'windshielding_1d']


def get_crs_info(srid):
    # Create temporary spatialite to find crs unit and projection
    conn = sqlite3.connect(":memory:")
    conn.enable_load_extension(True)
    conn.load_extension("mod_spatialite")
    # Initialite spatialite without any meta data
    conn.execute("SELECT InitSpatialMetaData(1, 'NONE');")
    # Add CRS
    success = conn.execute(f"SELECT InsertEpsgSrid({srid})").fetchone()[0]
    if not success:
        raise InvalidSRIDException(srid, "the supplied epsg_code is invalid")
    # retrieve units and is_projected
    unit = conn.execute(f'SELECT SridGetUnit({srid})').fetchone()[0]
    is_projected = conn.execute(f'SELECT SridIsProjected({srid})').fetchone()[0]
    return unit, is_projected


def get_model_srid() -> int:
    # Note: this will not work for models which are allowed to have no CRS (no geometries)
    conn = op.get_bind()
    srid_str = conn.execute(sa.text("SELECT epsg_code FROM model_settings")).fetchone()
    if srid_str is None or srid_str[0] is None:
        if not has_geom():
            return None
        raise InvalidSRIDException(None, "no epsg_code is defined")
    try:
        srid = int(srid_str[0])
    except TypeError:
        raise InvalidSRIDException(srid_str[0], "the epsg_code must be an integer")
    unit, is_projected = get_crs_info(srid)
    if unit != "metre":
        raise InvalidSRIDException(srid, f"the CRS must be in meters, not {unit}")
    if not is_projected:
        raise InvalidSRIDException(srid, "the CRS must be in projected")
    return srid


def get_geom_type(table_name, geo_col_name):
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    for col in columns:
        if col[1] == geo_col_name:
            return col[2]


def add_geometry_column(table: str, name: str, srid: int, geometry_type: str):
    # Adding geometry columns via alembic doesn't work
    query = (
        f"SELECT AddGeometryColumn('{table}', '{name}', {srid}, '{geometry_type}', 'XY', 1);")
    op.execute(sa.text(query))


def transform_column(table_name, srid):
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    # get all column names and types
    skip_cols = ['id', 'geom']
    col_names = [col[1] for col in columns if col[1] not in skip_cols]
    col_types = [col[2] for col in columns if col[1] not in skip_cols]
    # Create temporary table
    temp_table_name = f'_temp_230_{table_name}_{uuid.uuid4().hex}'
    # Create new table, insert data, drop original and rename temp to table_name
    col_str = ','.join(['id INTEGER PRIMARY KEY NOT NULL'] + [f'{cname} {ctype}' for cname, ctype in
                                                              zip(col_names, col_types)])
    query = op.execute(sa.text(f"CREATE TABLE {temp_table_name} ({col_str});"))
    # Add geometry column with new srid!
    geom_type = get_geom_type(table_name, 'geom')
    add_geometry_column(temp_table_name, 'geom', srid, geom_type)
    # Copy transformed geometry and other columns to temp table
    col_str = ','.join(['id'] + col_names)
    query = op.execute(sa.text(f"""
        INSERT INTO {temp_table_name} ({col_str}, geom) 
        SELECT {col_str}, ST_Transform(geom, {srid}) AS geom FROM {table_name}
        """))
    # Discard geometry column in old table
    op.execute(sa.text(f"SELECT DiscardGeometryColumn('{table_name}', 'geom')"))
    op.execute(sa.text(f"SELECT DiscardGeometryColumn('{temp_table_name}', 'geom')"))
    # Remove old table
    op.execute(sa.text(f"DROP TABLE '{table_name}'"))
    # Rename temp table
    op.execute(sa.text(f"ALTER TABLE '{temp_table_name}' RENAME TO '{table_name}';"))
    # Recover geometry stuff
    # This gives a bunch of warnings but seems to be needed to fix spatialite stuff
    op.execute(sa.text(f"SELECT RecoverGeometryColumn('{table_name}', "
                       f"'geom', {srid}, '{geom_type}', 'XY')"))
    op.execute(sa.text(f"SELECT CreateSpatialIndex('{table_name}', 'geom')"))
    op.execute(sa.text(f"SELECT RecoverSpatialIndex('{table_name}', 'geom')"))


def prep_spatialite(srid: int):
    conn = op.get_bind()
    has_srid = conn.execute(sa.text(f'SELECT COUNT(*) FROM spatial_ref_sys WHERE srid = {srid};')).fetchone()[0] > 0
    if not has_srid:
        conn.execute(sa.text(f"InsertEpsgSrid({srid})"))


def has_geom():
    connection = op.get_bind()
    has_data = [connection.execute(sa.text(f'SELECT COUNT(*) FROM {table}')).fetchone()[0] > 0 for table in GEOM_TABLES]
    return any(has_data)


def upgrade():
    # retrieve srid from model settings
    # raise exception if there is no srid, or if the srid is not valid
    srid = get_model_srid()
    if srid is None:
        print('Model without geometries and epsg code, we need to think about this')
        return
    # prepare spatialite databases
    prep_spatialite(srid)
    # transform all geometries
    for table_name in GEOM_TABLES:
        transform_column(table_name, srid)
    # remove crs from model_settings
    with op.batch_alter_table('model_settings') as batch_op:
        batch_op.drop_column('epsg_code')


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")
